import logging
import urllib.error
import urllib.request
from random import sample
from tempfile import NamedTemporaryFile
from typing import Union, Tuple

import numpy as np
from numpy import int8
from numpy import ndarray
from numpy.random import RandomState

from .simulation import Simulation

BIT_TYPE = int8

logger = logging.getLogger(__name__)


def random_inputs(n: int, N: int, seed: int) -> ndarray:
    r"""
    Generates :math:`N` uniformly random challenges of length `n` and returns them as a `numpy.ndarray` of shape
    :math:`(N, n)`. The randomness is based on the provided ``seed``.

    .. note::
        pypuf uses :math:`\{-1,1\}` to represent bit values for both challenges and responses.
        To convert from :math:`\{-1,1\}` notation to the more traditional :math:`\{0,1\}` encoding, use
        ``x = (1 - x) // 2``. For the reverse conversion, use ``x = 1 - 2*x``

        >>> import numpy as np
        >>> import pypuf.io
        >>> challenges = pypuf.io.random_inputs(n=64, N=10, seed=1)
        >>> np.unique(challenges)
        array([-1,  1], dtype=int8)
        >>> challenges01 = (1 - challenges) // 2
        >>> np.unique(challenges01)
        array([0, 1], dtype=int8)
        >>> challenges11 = 1 - 2 * challenges01
        >>> (challenges11 == challenges).all()
        True

    """
    return 2 * RandomState(seed).randint(0, 2, (N, n), dtype=BIT_TYPE) - 1


class ChallengeInformationSet:
    """
    Contains information about the behavior of a PUF token when queried with challenges.
    """
    # TODO fix type annotation when removing python 3.7 support

    def __init__(self, challenges: ndarray, information: ndarray) -> None:
        if challenges.shape[0] != information.shape[0]:
            raise ValueError('Must supply an equal number of challenges and information about these challenges.')
        self.challenges = challenges
        self.information = information
        self.N = len(self.challenges)

    @property
    def challenge_length(self) -> int:
        """
        The length :math:`n` of the recorded challenges.
        """
        return self.challenges.shape[1]

    def __len__(self) -> int:
        return self.challenges.shape[0]

    def __getitem__(self, item: Union[slice, int]) -> Union[Tuple[ndarray, ndarray], object]:
        if isinstance(item, int):
            return self.challenges[item], self.information[item]
        else:
            return self.__class__(self.challenges[item], self.information[item])

    def __eq__(self, other: object) -> bool:
        return (self.challenges == other.challenges).all() and (self.information == other.information).all()

    def random_subset(self, N: Union[int, float]) -> object:
        if N < 1:
            N = int(self.N * N)
        return self[sample(range(self.N), N)]

    def block_subset(self, i: int, total: int) -> object:
        return self[int(i / total * self.N):int((i + 1) / total * self.N)]

    def save(self, f: str) -> None:
        """
        Saves the CRPs to the given file ``f``.
        """
        np.savez_compressed(f, challenges=self.challenges, information=self.information)

    @classmethod
    def load(cls, f: str) -> object:
        """
        Loads CRPs from the given file ``f``.
        """
        return cls(*cls._load(f))

    @staticmethod
    def _load(f: str) -> Tuple[ndarray, ndarray]:
        data = np.load(f)
        challenges = data['challenges']
        information = data['information'] if 'information' in data else data['responses']
        return challenges, information

    def __repr__(self) -> str:
        return f"<{len(self)} CRPs with challenge length {self.challenge_length}>"


class ChallengeResponseSet(ChallengeInformationSet):

    @classmethod
    def from_simulation(cls, instance: Simulation, N: int, seed: int, r: int = 1) -> ChallengeInformationSet:
        challenges = random_inputs(instance.challenge_length, N, seed)
        crp_set = cls(
            challenges=challenges,
            responses=instance.r_eval(r, challenges)
        )
        crp_set.instance = instance
        return crp_set

    def __init__(self, challenges: ndarray, responses: ndarray) -> None:
        """
        Create a challenge-response object containing the given challenges and responses of a PUF token.

        :param challenges: Challenges to the PUF token organized as array of shape :math:`(N, n)`, where :math:`N`
            is the number of challenges and :math:`n` is the challenge length.
        :type challenges: `numpy.ndarray`
        :param responses: Responses of the PUF token organized as array of shape :math:`(N, m, r)`, where :math:`N`
            is the number of challenges, :math:`m` is the response length, and :math:`r` is the number of measurements
            per challenge. The last axis is optional.
        :type responses: `numpy.ndarray`
        """
        super().__init__(challenges, responses)
        if len(responses.shape) == 1:
            responses = responses.reshape(responses.shape + (1, 1))
        elif len(responses.shape) == 2:
            responses = responses.reshape(responses.shape + (1, ))
        self.responses = responses

    @property
    def response_length(self) -> int:
        """
        The length :math:`m` of the recorded responses.
        """
        return self.responses.shape[1]

    def __repr__(self) -> str:
        return f"<{len(self)} CRPs with challenge length {self.challenge_length} and response " \
               f"length {self.response_length}, each response measured {self.responses.shape[2]} time(s)>"


class ChallengeReliabilitySet(ChallengeInformationSet):

    @classmethod
    def from_simulation(cls, instance: Simulation, N: int, seed: int, r: int = 5) -> ChallengeInformationSet:
        # noinspection PyTypeChecker
        return cls.from_challenge_response_set(ChallengeResponseSet.from_simulation(instance, N, seed, r))

    @classmethod
    def from_challenge_response_set(cls, crp_set: ChallengeResponseSet) -> ChallengeInformationSet:
        return cls(
            challenges=crp_set.challenges,
            reliabilities=np.average(crp_set.responses, axis=-1),
        )

    def __init__(self, challenges: ndarray, reliabilities: ndarray) -> None:
        super().__init__(challenges, reliabilities)
        self.reliabilities = reliabilities


fetchable = object()


class LazyCRPs(ChallengeResponseSet):
    challenges = fetchable
    responses = fetchable
    N = fetchable

    def __init__(self, url: str) -> None:
        self.url = url
        self._fetched = False

    def __getattribute__(self, name: str) -> object:
        attr = super().__getattribute__(name)
        if attr is fetchable:
            self.fetch()
            return super().__getattribute__(name)
        else:
            return attr

    def fetch(self) -> None:
        try:
            try:
                size = int(urllib.request.urlopen(
                    urllib.request.Request(url=self.url, method="HEAD")
                ).headers.get('Content-Length'))
                size = f"{size/1024**2:.1f}MiB"
            except (ValueError, urllib.error.HTTPError, urllib.error.URLError):
                size = 'unknown size'

            logger.warning(f"Fetching CRPs ({size}) from {self.url} ...")
            self._fetched = True
            with NamedTemporaryFile() as f:
                urllib.request.urlretrieve(self.url, f.name)
                self.challenges, self.information = self._load(f.name)
                self.responses = self.information
                self.N = len(self.challenges)
        except Exception as e:
            self._fetched = False
            raise e

    def __repr__(self) -> str:
        return super().__repr__() if self._fetched else f"<CRP Set available from {self.url}, not fetched yet>"


class MTZAA20:
    xor_arbiter_puf_4_xor = LazyCRPs(
        "https://zenodo.org/record/5215875/files/MTZAA20_4XOR_64bit_LUT_2239B_attacking_1M.txt.npz?download=1")
    xor_arbiter_puf_5_xor = LazyCRPs(
        "https://zenodo.org/record/5215875/files/MTZAA20_5XOR_64bit_LUT_2239B_attacking_1M.txt.npz?download=1")
    xor_arbiter_puf_6_xor = LazyCRPs(
        "https://zenodo.org/record/5215875/files/MTZAA20_6XOR_64bit_LUT_2239B_attacking_1M.txt.npz?download=1")
    xor_arbiter_puf_7_xor = LazyCRPs(
        "https://zenodo.org/record/5215875/files/MTZAA20_7XOR_64bit_LUT_2239B_attacking_5M.txt.npz?download=1")
    xor_arbiter_puf_8_xor = LazyCRPs(
        "https://zenodo.org/record/5215875/files/MTZAA20_8XOR_64bit_LUT_2239B_attacking_5M.txt.npz?download=1")
    xor_arbiter_puf_9_xor = LazyCRPs(
        "https://zenodo.org/record/5215875/files/MTZAA20_9XOR_64bit_LUT_2239B_attacking_5M.txt.npz?download=1")


class AM21:
    arbiter_puf_bottom_0 = LazyCRPs("https://zenodo.org/record/5221305/files/AM21_b0.npz?download=1")
    arbiter_puf_bottom_1 = LazyCRPs("https://zenodo.org/record/5221305/files/AM21_b1.npz?download=1")
    arbiter_puf_bottom_2 = LazyCRPs("https://zenodo.org/record/5221305/files/AM21_b2.npz?download=1")
    arbiter_puf_bottom_3 = LazyCRPs("https://zenodo.org/record/5221305/files/AM21_b3.npz?download=1")
    arbiter_puf_bottom_4 = LazyCRPs("https://zenodo.org/record/5221305/files/AM21_b4.npz?download=1")
    arbiter_puf_top = LazyCRPs("https://zenodo.org/record/5221305/files/AM21_top.npz?download=1")
    interpose_puf = LazyCRPs("https://zenodo.org/record/5221305/files/AM21_ipuf.npz?download=1")


class CCPG21:
    hbn_board_1 = LazyCRPs("https://zenodo.org/record/5526722/files/CCPG21_board1.npz?download=1")
    hbn_board_2 = LazyCRPs("https://zenodo.org/record/5526722/files/CCPG21_board2.npz?download=1")
    hbn_board_3 = LazyCRPs("https://zenodo.org/record/5526722/files/CCPG21_board3.npz?download=1")
    hbn_board_4 = LazyCRPs("https://zenodo.org/record/5526722/files/CCPG21_board4.npz?download=1")
    hbn_board_5 = LazyCRPs("https://zenodo.org/record/5526722/files/CCPG21_board5.npz?download=1")
    hbn_board_6 = LazyCRPs("https://zenodo.org/record/5526722/files/CCPG21_board6.npz?download=1")
    hbn_board_7 = LazyCRPs("https://zenodo.org/record/5526722/files/CCPG21_board7.npz?download=1")
    hbn_board_8 = LazyCRPs("https://zenodo.org/record/5526722/files/CCPG21_board8.npz?download=1")



from typing import Callable, Optional

import numpy as np

from .base import OfflineAttack
from ..io import ChallengeResponseSet
from ..simulation.base import Simulation


class LinearMapSimulation(Simulation):

    @staticmethod
    def postprocessing_id(responses: np.ndarray) -> np.ndarray:
        return responses

    @staticmethod
    def postprocessing_threshold(responses: np.ndarray) -> np.ndarray:
        return np.sign(responses)

    def __init__(self, linear_map: np.ndarray, challenge_length: int,
                 feature_map: Optional[Callable[[np.ndarray], np.ndarray]] = None,
                 postprocessing: Optional[Callable[[np.ndarray], np.ndarray]] = None) -> None:
        super().__init__()
        self.map = linear_map
        self._challenge_length = challenge_length
        self.feature_map = feature_map or (lambda x: x)
        self.postprocessing = postprocessing or self.postprocessing_id

    @property
    def challenge_length(self) -> int:
        return self._challenge_length

    @property
    def response_length(self) -> int:
        return self.map.shape[1]

    def eval(self, challenges: np.ndarray) -> np.ndarray:
        return self.postprocessing(self.feature_map(challenges) @ self.map)


class LeastSquaresRegression(OfflineAttack):

    @staticmethod
    def feature_map_linear(challenges: np.ndarray) -> np.ndarray:
        return challenges

    @staticmethod
    def feature_map_optical_pufs_reloaded(challenges: np.ndarray) -> np.ndarray:
        """
        Computes features of an optical PUF token using all ordered pairs of challenge bits [RHUWDFJ13]_.
        An optical system may be linear in these features.

        .. note::
            This representation is redundant since it treats ordered paris of challenge bits are distinct.
            Actually, only unordered pairs of bits should be treated as distinct. For applications, use
            the function :meth:`feature_map_optical_pufs_reloaded_improved
            <pypuf.attack.LeastSquaresRegression.feature_map_optical_pufs_reloaded_improved>`,
            which achieves the same with half the number of features.

        :param challenges: array of shape :math:`(N, n)` representing challenges to the optical PUF.
        :return: array of shape :math:`(N, n^2)`, which, for each challenge, contains the flattened dyadic product of
            the challenge with itself.
        """
        beta = np.einsum("...i,...j->...ij", challenges, challenges)
        return beta.reshape(beta.shape[:-2] + (-1,))

    @staticmethod
    def feature_map_optical_pufs_reloaded_improved(challenges: np.ndarray) -> np.ndarray:
        r"""
        Computes features of an optical PUF token using all unordered pairs of challenge bits [RHUWDFJ13]_.
        An optical system may be linear in these features.

        :param challenges: 2d array of shape :math:`(N, n)` representing `N` challenges of length :math:`n`.
        :return: array of shape :math:`(N, \frac{n \cdot (n + 1)}{2})`. The result `return[i]` consists of all products
            of unordered pairs taken from `challenges[i]`, which has shape `(N,)`.

        >>> import numpy as np
        >>> import pypuf.attack
        >>> challenges = np.array([[2, 3, 5], [1, 0, 1]])  # non-binary numbers for illustration only.
        >>> pypuf.attack.LeastSquaresRegression.feature_map_optical_pufs_reloaded_improved(challenges)
        array([[ 4,  6, 10,  9, 15, 25],
               [ 1,  0,  1,  0,  0,  1]])
        """
        n = challenges.shape[1]
        idx = np.triu_indices(n)
        return np.einsum("...i,...j->...ij", challenges, challenges)[:, idx[0], idx[1]]

    def __init__(self, crps: ChallengeResponseSet,
                 feature_map: Optional[Callable[[np.ndarray], np.ndarray]] = None) -> None:
        super().__init__(crps)
        self.crps = crps
        self.feature_map = feature_map or (lambda x: x)
        self._model = None

    def fit(self) -> Simulation:
        features = self.feature_map(self.crps.challenges)
        # TODO warn if more than one measurement
        linear_map = np.linalg.pinv(features) @ self.crps.responses[:, :, 0]
        self._model = LinearMapSimulation(linear_map, self.crps.challenge_length, self.feature_map)
        return self.model

    @property
    def model(self) -> Simulation:
        return self._model



import numpy as np
from more_itertools import distinct_permutations
from scipy.special import comb

from .base import OfflineAttack
from ..io import ChallengeResponseSet
from ..simulation.base import Simulation

_low_degree_set_cache = {}


def to_low_degree_chisx(X: np.ndarray, deg: int) -> np.ndarray:
    """
    Given a list of inputs `X` of :math:`n` bits each, and the maximum parity function degree `deg`,
    computes the values of all parity functions up to degree `deg` for each of the inputs given in `X`.

    >>> import numpy as np
    >>> X = np.array([[1, 1, 1, 1], [-1, -1, -1, -1]])
    >>> to_low_degree_chisx(X, deg=1)
    array([[ 1,  1,  1,  1,  1],
           [ 1, -1, -1, -1, -1]])

    :param X: List of :math:`N` inputs with :math:`n` bits each, given as numpy array of shape :math:`(N, n)`.
    :param deg: Maximum degree of the parity functions.
    :return: List of :math:`N` vectors containing the values of the parity functions for each given input.
        Represented by a numpy array of shape :math:`(N, k)`, where :math:`k` is the number of parity functions of
        degree up to `deg`.
    """
    return 1 - 2 * (((1 - X) // 2) @ low_degree_sets(X.shape[1], deg).T % 2)


def low_degree_sets(n: int, deg: int = 1) -> np.ndarray:
    """
    Returns a list of parity functions of :math:`n` bit of degree up to `deg`.
    The parity functions are represented by a vector of length :math:`n` indicating of which bits the parity is
    computed.

    >>> low_degree_sets(4, deg=1)
    array([[0, 0, 0, 0],
           [0, 0, 0, 1],
           [0, 0, 1, 0],
           [0, 1, 0, 0],
           [1, 0, 0, 0]], dtype=int8)

    Computation can be expensive, results are cached.

    :param n: Number of bits to compute parities from.
    :param deg: Maximum degree of the parity function, i.e. up to how many bits are xored.
    :return: List of parity functions represented by vectors indicating bit positions, formatted as numpy array of
        shape `(num,n)`, where `num` is the number of parity functions.
    """
    if _low_degree_set_cache.get((n, deg)) is not None:
        return _low_degree_set_cache.get((n, deg))

    l = sum(comb(N=n, k=d) for d in range(deg + 1))
    ss = np.empty(shape=(int(l), n), dtype=np.int8)

    idx = 0
    for d in range(deg + 1):
        dl = int(comb(N=n, k=d))
        ss[idx:idx + dl] = np.array(list(distinct_permutations([1] * d + [0] * (n - d), r=n)))
        idx += dl

    _low_degree_set_cache[(n, deg)] = ss
    return ss


class FourierSimulation(Simulation):
    r"""
    A function :math:`f: \{0, 1\}^n \to \mathbb{R}`, defined by its Fourier coefficients.
    """

    def __init__(self, expansion: np.ndarray, deg: int, challenge_length: int, boolean: bool = True) -> None:
        super().__init__()
        self.expansion = expansion
        self.deg = deg
        self._challenge_length = challenge_length
        self.boolean = boolean

    @property
    def challenge_length(self) -> int:
        return self._challenge_length

    @property
    def response_length(self) -> int:
        return self.expansion.shape[0]

    def eval(self, challenges: np.ndarray) -> np.ndarray:
        val = self.eval_chisx(to_low_degree_chisx(challenges, self.deg))
        return np.sign(val) if self.boolean else val

    def eval_chisx(self, chisx: np.ndarray) -> np.ndarray:
        N, f = chisx.shape  # N number of inputs for prediction, f number of Fourier coefficients used for prediction
        m = self.expansion.shape[0]  # number of predictions for each input

        # compute predictions
        Yp = np.empty(shape=(N, m), dtype=np.float32)
        for i in range(m):
            assert self.expansion[i].shape == (f,)
            assert chisx.T.shape == (f, N)
            assert Yp[:, i].shape == (N,)
            Yp[:, i] = self.expansion[i] @ chisx.T
        return Yp


class LMNAttack(OfflineAttack):

    def __init__(self, crps: ChallengeResponseSet, deg: int = 1) -> None:
        r"""
        Given a list of function values of a function :math:`f: \{-1,1\}^n \to \mathbb{R}`, an approximation of the
        Fourier spectrum of the underlying function can be computed.

        The approximation is guaranteed to be correct, if the list contains all function values and the degree equals
        :math:`n`, shown here using the Boolean AND function:

        >>> import numpy as np
        >>> import pypuf.io, pypuf.attack
        >>> challenges = 1 - 2 * np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
        >>> responses_and = 1 - 2 * np.array([[0], [0], [0], [1]])
        >>> crps_and = pypuf.io.ChallengeResponseSet(challenges, responses_and)
        >>> (1 - pypuf.attack.LMNAttack(crps_and, 2).fit().eval(challenges)) // 2
        array([[0.],
               [0.],
               [0.],
               [1.]], dtype=float32)

        If additionally the responses are from :math:`\{-1,1\}`, then the sum of squares of the Fourier coefficients
        equals 1, as illustrated here using the majority vote function:

        >>> challenges = 1 - 2 * np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 1], [1, 0, 0], [1, 0, 1], [1, 1, 0], [1, 1, 1]])
        >>> responses_maj = 1 - 2 * np.array([[0], [0], [0], [1], [0], [1], [1], [1]])
        >>> crps_maj = pypuf.io.ChallengeResponseSet(challenges, responses_maj)
        >>> exp = pypuf.attack.LMNAttack(crps_maj, 3).fit().expansion
        >>> exp
        array([[ 0. ,  0.5,  0.5,  0.5,  0. ,  0. ,  0. , -0.5]], dtype=float32)
        >>> np.sum(exp**2)
        1.0
        """
        super().__init__(crps)
        self.deg = deg
        self.prediction_shape = None
        self._model = None

    def fit(self) -> FourierSimulation:
        return self.fit_chisx(
            to_low_degree_chisx(self.crps.challenges, self.deg),
            self.crps.responses,
            self.crps.challenge_length,
        )

    def fit_chisx(self, chisx: np.ndarray, Y: np.ndarray, challenge_length: int) -> FourierSimulation:
        # preserve original response shape and flatten after the first axis
        self.prediction_shape = Y.shape[1:]
        Y = Y.reshape((Y.shape[0], -1))

        f = chisx.shape[1]  # number of Fourier coefficients to be approximated
        m = Y.shape[1]  # number of expansions to be computed
        N = Y.shape[0]  # number of examples
        assert N == chisx.shape[0]

        # compute expansions
        expansion = np.empty(shape=(m, f), dtype=np.float32)
        for i in range(m):
            assert Y[:, i].shape == (N,)
            assert chisx.shape == (N, f)
            assert expansion[i].shape == (f,)
            expansion[i] = Y[:, i] @ chisx / f

        # create model
        self._model = FourierSimulation(expansion, self.deg, challenge_length)
        return self.model

    @property
    def model(self) -> Simulation:
        return self._model



from ..io import random_inputs
from ..simulation import Simulation
from numpy import ndarray
import numpy as np


def flip_ith_bit(challenges: ndarray, i: int) -> ndarray:
    challenges = challenges[:]
    challenges[:, i] *= -1
    return challenges


def influence(puf: Simulation, i: int, seed: int, N: int = 1000) -> float:
    r"""
    Approximates the influence of the :math:`i`-th input bit on the output of the given PUF simulation.

    For a Boolean function :math:`f:\{-1,1\}^n\to\{-1,1\}`, the influence of the :math:`i`-th input coordinate is
    defined as the probability that the output of the function changes when the :math:`i`-th input coordinate is
    changed, i.e.

    .. math::
        \mathrm{Inf}_i(f) = \Pr_x \left[ f(x) \neq f(x^{\oplus i}) \right],

    where :math:`x^{\oplus i}` is the same as :math:`x`, but with the :math:`i`-th input coordinate flipped.

    The value of :math:`\mathrm{Inf}_i(f)` is approximated on a sample of uniform random inputs to :math:`f`.

    Ideally, all input bits have an influence of approximately 50% on the response of a PUF function.
    In contrast, the Arbiter PUF is known to have input bits with extremely high and low influences:

    >>> import pypuf.simulation, pypuf.metrics
    >>> puf = pypuf.simulation.ArbiterPUF(n=128, seed=1)
    >>> pypuf.metrics.influence(puf, i=0, seed=2)
    0.03
    >>> pypuf.metrics.influence(puf, i=127, seed=3)
    0.971

    The Lightweight Secure PUF [MKP08]_ removed this flaw of the Arbiter PUF, but did not succeed in created a more
    secure PUF [WBMS19]_.

    >>> pypuf.metrics.influence(pypuf.simulation.LightweightSecurePUF(n=128, k=1, seed=1), i=0, seed=2)
    0.447

    :param puf: Function :math:`f`.
    :type puf: :class:`pypuf.simulation.Simulation`
    :param i: The (zero-based) index of the coordinate whose influence will be approximated.
    :type i: ``int``
    :param seed: Determines the seed for the PRNG that generates the uniform random inputs to :math:`f`.
    :type seed: ``int``
    :param N: The number of uniform random inputs that is used to approximate the influence.
    :type N: ``int``
    :return: An approximation of the influence of the :math:`i`-th input coordinate on the output of :math:`f`, i.e.
        :math:`\mathrm{Inf}_i(f)`.
    :rtype: ``float``
    """
    n = puf.challenge_length
    c = random_inputs(n=n, N=N, seed=seed)
    return np.mean(puf.eval(c) != puf.eval(flip_ith_bit(c, i)))


def total_influence(puf: Simulation, seed: int, N: int = 1000) -> float:
    r"""
    The total influence, also known as average sensitivity [ODon14]_, is the sum of all coordinate-wise influences as
    computed by :func:`influence`, i.e.

    .. math::
        \mathrm{I}(f) = \sum_{i=1}^n \Pr_x \left[ f(x) \neq f(x^{\oplus i}) \right].

    We hence have :math:`0 \leq \mathrm{I}(f) \leq n`.

    The total influence can give information about the membership of the function :math:`f` in certain classes and
    is used in PUFMeter to analyze PUFs with respect to modeling attacks [GFS19]_. Low values of total influence are
    concerning as they indicate that a modeling attack using the LMN algorithm may be successful.

    As an example, the total influence of the :class:`pypuf.simulation.ArbiterPUF` and
    :class:`pypuf.simulation.PermutationPUF` can be computed as follows:

    >>> import pypuf.simulation, pypuf.metrics
    >>> pypuf.metrics.total_influence(pypuf.simulation.ArbiterPUF(n=64, seed=1), seed=2)
    19.358
    >>> pypuf.metrics.total_influence(pypuf.simulation.PermutationPUF(n=64, k=1, seed=1), seed=2)
    32.263...

    :param puf: Function :math:`f`.
    :type puf: :class:`pypuf.simulation.Simulation`
    :param seed: Determines the seed for the PRNG that generates the uniform random inputs to :math:`f`.
    :type seed: ``int``
    :param N: The number of uniform random inputs that is used to approximate the influence.
    :type N: ``int``
    :return: An approximation of the total influence of :math:`f`, i.e. :math:`\mathrm{I}_i(f)`.
    :rtype: ``float``
    """
    n = puf.challenge_length
    c = random_inputs(n=n, N=N, seed=seed)
    r = puf.eval(c)  # original responses
    return sum(
        np.mean(r != puf.eval(flip_ith_bit(c, i)))
        for i in range(n)
    )


def noise_sensitivity(puf: Simulation, eps: float, seed: int, N: int = 1000) -> float:
    r"""
    The noise sensitivity captures how a function reacts to noisy inputs. Note that noise here refers to flipped
    *input bits*, i.e. is different from the usual notion of noise in the context of PUFs. Formally, the noise
    sensitivity of a Boolean function :math:`f` at noise-level :math:`\varepsilon` is defined as the probability that
    the output flips if each input bit is flipped with probability :math:`\varepsilon` (independently). I.e.,

    .. math::
        \mathrm{NS}_\varepsilon(f) = \Pr_x \left[ f(x) \neq f(x') \right],

    where :math:`x'` is a copy of :math:`x`, but each bit is flipped independently with probability :math:`\varepsilon`.

    The noise sensitivity is approximated using a uniform random sample of inputs and corresponding
    with-probability-:math:`\varepsilon`-flipped inputs to :math:`f`.

    A low noise sensitivity of :math:`f` indicates that further testing is required, as :math:`f` may be close to a
    junta and thus susceptible to a modeling attack [GFS19]_.

    Usually, it is interesting to compute the noise sensitivity for small :math:`\varepsilon`:

    >>> import pypuf.simulation, pypuf.metrics
    >>> puf = pypuf.simulation.ArbiterPUF(n=64, seed=1)
    >>> pypuf.metrics.noise_sensitivity(puf, eps=.01, seed=2)
    0.216

    :param puf: Function :math:`f`.
    :type puf: :class:`pypuf.simulation.Simulation`
    :param eps: Noise-level :math:`\varepsilon` of the inputs.
    :type eps: ``float``
    :param seed: Determines the seed for the PRNG that generates the inputs to :math:`f`.
    :type seed: ``int``
    :param N: The number of uniform random inputs that is used to approximate the noise sensitivity.
    :type N: ``int``
    :return: The noise sensitivity of :math:`f` at noise level :math:`\varepsilon`, i.e.
        :math:`\mathrm{NS}_\varepsilon(f)`.
    :rtype: ``float``
    """
    n = puf.challenge_length
    challenges = random_inputs(n=n, N=N, seed=seed)
    flip = np.random.default_rng(seed).choice([-1, 1], size=(N, n), p=[eps, 1 - eps])
    return np.mean(puf.eval(challenges) != puf.eval(challenges * flip))



from .common import reliability_data, reliability, uniqueness_data, uniqueness, similarity_data, accuracy, similarity, \
    bias, bias_data, correlation, correlation_data
from .fourier import total_influence, noise_sensitivity, influence



from numpy import ndarray
import numpy as np

from ..simulation import Simulation


class IntegratedOpticalPUF(Simulation):
    r"""
    Very basic simulation of an integrated optical PUF [RHUWDFJ13]_.
    This simulation is derived from the corresponding successful modeling attack [RHUWDFJ13]_.

    To compute a response to a given challenge :math:`c`, the simulation evaluates the value of

    .. math::
        \left| c \cdot A \cdot e^{\phi i} \right|^2,

    where :math:`A \in \mathbb{R}^{n \times m}` and :math:`\phi \in \mathbb{R}^{n \times m}`; :math:`n` is the
    challenge length, :math:`m` the response length.

    By default, the values for :math:`A` are chosen mutually independent and uniformly random in :math:`[0,1)`;
    the values for :math:`\phi` are chosen mutually independent and uniformly random in :math:`[0, 2\pi)`.

    .. warning::
        This simulation only gives a very rough, idealized idea of how integrated optical PUFs may behave.
        In particular, the :meth:`pypuf.metrics.uniqueness` may be quite different from the behavior real-world
        implementations. Also, the behavior of response pixels is mutually independent, which differs from behavior
        reported in the literature [RHUWDFJ13]_.
    """

    def __init__(self, n: int, m: int, seed: int) -> None:
        """
        Initializes a simulation for an Integrated Optical PUF with :math:`n` challenge bits, :math:`m` response bits,
        with randomness based on the given ``seed``.
        """
        super().__init__()
        self.n = n
        self.m = m

        A = np.random.default_rng(self.seed(f"intg. optical PUF {seed} A")).uniform(0, 1, size=(n, m))
        phi = np.random.default_rng(self.seed(f"intg. optical PUF {seed} phi")).uniform(0, 2 * np.pi, size=(n, m))
        self.T = A * np.exp(1j * phi)

    @property
    def challenge_length(self) -> int:
        return self.n

    @property
    def response_length(self) -> int:
        return self.m

    def eval(self, challenges: ndarray) -> ndarray:
        return np.abs(challenges @ self.T)**2
